/* file yasep/JScore/group_asu.js
created 2005-06-11 by whygee@f-cpu.org
2007-04-11 : moved from vsp_eu.js to group_asu.js, stripped ADD1, ADDC1, SUB1, SUBB1
2007-04-15 : added the USE_ flags
2007-11-05 : rebranding to YASEP
2008-11-30 :
 * removed ADDS3 & SUBS3
 * moved UMIN/UMAX from group_mov to here
 * 16-bit actions added
2009-04-07 : replaced ADDS1/ADDS2/SUBS1/SUBS2 with COMPU/COMPS/SMIN/SMAX
2009-07-27 : Y
2009-07-29 : FORM_iR, SUB has Y.FLAG_SWAPDEST
2009-08-04 : removed USE_IMM16 and READ_SRC1orIMM
   missing carry flag generation
2009-08-08 : new version of the computation, closer to how VHDL is coded.
2009-08-10 : removed SWAPDEST

This file implements the "ASU" (Add/Sub) Execution Unit

dependency : group_eu.js and all its parents
*/
Y.prefix=Y.GROUP_ASU;

function carry_16(a,b){
  return (((a^F4)+1)+b)>>16 // ^ 0x10000;
 // note : The ASU does not do the final XOR on the carry bit
 // so the borrow is reversed for SUB
}

function carry_ADD32(a,b) {
  var res=(b & F4)+(a & F4);
  res=((b>>16)& F4)+((a>>16)& F4)+((res>>16)&1);
  return (res>>16)&1;
}

function carry_SUB32(a,b){
  return carry_ADD32((a^F8)+1,b);
}

Y.NewOpcode("ADD",
  function(s){
    s.result = (s.snd+s.si4)&F8;
    s.carry  = carry_ADD32(s.snd,s.si4);
  },
  function(s){
    s.result = s.snd+s.si4;
    s.carry  = s.result >> 16;
    s.result &= F4;
  },
  Y.FORM_RR|Y.FORM_IRR|Y.FORM_iR|Y.FORM_IR|Y.FORM_iRR|Y.FORM_RRR,
  Y.ALIAS_IRR|Y.CHANGE_CARRY, "Addition");

// this alias generates a SUB with Imm4/Imm16=0
Y.NewAlias("NEG", "SUB", Y.FORM_RR|Y.FORM_R,
  Y.ALIAS_RR|Y.WRITE_SRC1|Y.CHANGE_CARRY, "arithmetic negation");
Y.NewOpcode("SUB",
  function(s){
    s.result = (s.si4-s.snd)&F8;
    s.carry  = carry_SUB32(s.snd,s.si4);
  },
  function(s) {
    s.result = (s.si4 - s.snd)&F4;
    s.carry  = carry_16(s.snd,s.si4);
  },
  Y.FORM_RR|Y.FORM_IRR|Y.FORM_iR|Y.FORM_IR|Y.FORM_iRR|Y.FORM_RRR,
  Y.ALIAS_IRR|Y.CHANGE_CARRY, "Substraction");

Y.NewOpcode("CMPU",
  function(s){
    s.carry  = carry_SUB32(s.snd,s.si4);
  },
  function(s) {
    s.carry = carry_16(s.snd,s.si4);
  },
  Y.FORM_RR|Y.FORM_IR|Y.FORM_iR,
  Y.NO_CONDITION|Y.NO_WRITEBACK|Y.CHANGE_CARRY,
  "unsigned substract the 2 operands and set the LSB of NPC if borrow");

Y.NewOpcode("CMPS",
  function(s){
    s.carry  = carry_SUB32(s.snd^0x80000000,s.si4^0x80000000);
  },
  function(s) {
    s.carry = carry_16(s.snd^0x8000, s.si4^0x8000);
  },
  Y.FORM_RR|Y.FORM_IR|Y.FORM_iR,
  Y.NO_CONDITION|Y.NO_WRITEBACK|Y.CHANGE_CARRY,
  "signed substract the 2 operands and set the LSB of NPC if borrow");

Y.NewOpcode("UMIN",
  function(s){
    if (!carry_SUB32(s.snd,s.si4))
      s.result = s.si4;
  },
  function(s) {
    if (!carry_16(s.snd, s.si4))
      s.result = s.si4;
  },
  Y.FORM_RR|Y.FORM_IRR|Y.FORM_iR|Y.FORM_IR|Y.FORM_iRR|Y.FORM_RRR,
  Y.ABORT_WRITE|Y.ALIAS_IRR,
  "Set dest to imm16/si4 if snd's value is less (unsigned) than imm16/si4"
);

Y.NewOpcode("UMAX",
  function(s){
    if (carry_SUB32(s.snd,s.si4))
      s.result = s.si4;
  },
  function(s) {
    if (carry_16(s.snd, s.si4))
      s.result = s.si4;
  },
  Y.FORM_RR|Y.FORM_IRR|Y.FORM_iR|Y.FORM_IR|Y.FORM_iRR|Y.FORM_RRR,
  Y.ABORT_WRITE|Y.ALIAS_IRR,
  "Set dest to imm16/si4 if snd's value is greater (unsigned) than imm16/si4"
);


Y.NewOpcode("SMIN",
  function(s){
    if (!carry_SUB32(s.snd^0x80000000,s.si4^0x80000000))
      s.result = s.si4;
  },
  function(s) {
    if (!carry_16(s.snd^0x8000, s.si4^0x8000))
      s.result = s.si4;
  },
  Y.FORM_RR|Y.FORM_IRR|Y.FORM_iR|Y.FORM_IR|Y.FORM_iRR|Y.FORM_RRR,
  Y.ABORT_WRITE|Y.ALIAS_IRR,
  "Set dest to imm16/si4 if snd's value is less (signed) than imm16/si4"
);

Y.NewOpcode("SMAX",
  function(s){
    if (carry_SUB32(s.snd^0x80000000,s.si4^0x80000000))
      s.result = s.si4;
  },
  function(s) {
    if (carry_16(s.snd^0x8000, s.si4^0x8000))
      s.result = s.si4;
  },
  Y.FORM_RR|Y.FORM_IRR|Y.FORM_iR|Y.FORM_IR|Y.FORM_iRR|Y.FORM_RRR,
  Y.ABORT_WRITE|Y.ALIAS_IRR,
  "Set dest to imm16/si4 if snd's value is greater (signed) than imm16/si4"
);
