/* File yasep/JScore/yasep_disasm.js
created 2006-09-04 by whygee@f-cpu.org
2006-11-10 (corrected an edition glitch ???)
2007-04-04 : update for the Dx pseudo-instructions
2007-04-09 : separation of forms and flags
2007-04-17 : another major update
2007-11-05 : rebranding to YASEP, support of the flags IMM16_5LSB and IMM16_8LSB
2009-07-30 : Y and others
2009-08-10 : rebuild !
2009-08-15 : RIR=>IRR, +decode_extended
2009-08-18 : factoring ? trying a new approach where the 5 main forms are separated,
              and 5 functions try to infer the aliases and short versions
2009-08-23 : working again

This file does the disassembly (decoding) of encoded instructions.
It transforms a number into a string and a size, for interactive debug and testing.

dependency:
 yasep/JScore/txt_int.js,
 (yasep/JSgui/yasep_messages.js+handling [define an empty function if undesired]),
 yasep/JScore/yasep_forms.js/flags/groups/fields
 yasep/JScore/yasep_opcodes.js+children (group_asu.js etc.)
*/

// constructor :
function Yasep_DISASM(){
 this.YASEP_TYPE=0; // public and accessible : no need of a setter/getter function

/* shared by the other functions :
  this.prefix=input & (Y.FIELD_IMMREG|Y.FIELD_EXTFORM);
  this.opcode=input &  Y.FIELD_OPCODE;
  this.input=input;
  this.instruction='';
  this.size_bytes=0;
  this.form=null;
  this.snd=(input & Y.FIELD_SND)>>Y.OFFSET_SND;
  this.si4=(input & Y.FIELD_SI4)>>Y.OFFSET_SI4;
  this.name=null;
*/


 // This piece of code is used both by short instructions
 // and long instructions with ignore (where the ? and X are added by the caller)
 this.decode_short=function(t,f,g){

// decode the Imm4 forms first
  var i4=t.si4;
  if (!(g & Y.IgnoreImmSign) && (i4 >= 8))
    i4 -= 16;

  if (t.input & Y.FIELD_EXTFORM){
    // aliases :
    if (t.si4==0){
      if (t.name=="SUB"){
        t.name="NEG";
        t.instruction=" "+Y.registers_names[t.snd];
        return 'FORM_R'
      }
      if (t.name=="ORN"){
        t.name="NOT";
        t.instruction=" "+Y.registers_names[t.snd];
        return 'FORM_R'
      }
    }
    if (f & Y.FORM_iR){
      t.instruction=" "+i4+" "+Y.registers_names[t.snd];
      return 'FORM_iR'
    }
    if(f & Y.FORM_Ri){
      t.instruction=" "+Y.registers_names[t.snd]+" "+i4;
      return 'FORM_Ri'
    }
  }
  if (f & Y.FORM_i){ // CRIT
    t.instruction=" "+i4;
    if (t.snd != 0)
      add_message('yellow', 'warning : garbage found in SND');
    return 'FORM_i'
  }

  // some ALONE are already filtered out but not the aliases :
  if ((t.input & 0xFFFE)==0) {
    t.name="NOP";
    return 'FORM_ALONE'
  }

  // RR :
  t.instruction=" "+Y.registers_names[t.si4]+" "+Y.registers_names[t.snd];
  return 'FORM_RR'
 }

 //////////////////////////// IRR ////////////////////////////
 this.decode_long=function(t,f,g){
  var l=(t.input>>16)&65535;
  if (!(g & Y.IgnoreImmSign) && (l >= 32768))
    l -= 65536;
   
  if ((f & Y.IGNORE_IMM16) // does it ignore the Imm16 field ?
  ||(t.input & 0xFFFE)==0) { // alias NOP
    if (f & Y.IGNORE_IMM16) {
      t.form='FORM_ALONE';
      if (l!=0)
        add_message('yellow', 'warning : garbage found in IMM16');
      if (t.snd != 0)
        add_message('yellow', 'warning : garbage found in SND');
      if (t.si4 != 0)
        add_message('yellow', 'warning : garbage found in SI4');
    }
    else
      return t.decode_short(t,f,g); // NOP ?
  }

  if (f&Y.FORM_RI) { // PUT
    if (t.si4 != 0)
      add_message('yellow', 'warning : garbage found in SI4');
    t.instruction=" "+Y.registers_names[t.snd]+" "+l;
    return 'FORM_RI'
  }   

  if (f&Y.FORM_IR) {
    if (g&Y.ALIAS_IRR){
      if (t.snd==t.si4) { // condensed writing
        t.instruction=" "+l+" "+Y.registers_names[t.snd]; // either si4 or snd, they are equal...
        return 'FORM_IR'
      }
    }
    else {
      if (g & Y.SWAP_IR){ // MOV, GET
        if (t.snd != 0)
          add_message('yellow', 'warning : garbage found in SND');
        t.instruction=" "+l+" "+Y.registers_names[t.si4];
      }
      else { // CMPU, CMPS
        if (t.si4 != 0)
          add_message('yellow', 'warning : garbage found in SI4');
        t.instruction=" "+l+" "+Y.registers_names[t.snd];
      }
      return 'FORM_IR'
    }
  }
  // default
  t.instruction=" "+l+" "+Y.registers_names[t.snd]+" "+Y.registers_names[t.si4];
  return 'FORM_IRR'
 }

 this.decode_extended=function(t,f,g){
  // decode the Imm4 forms first
  if (t.input & Y.FIELD_EXTIMM4) {
    var i4=t.si4;
    if (!(g & Y.IgnoreImmSign) && (i4 >= 8))
      i4 -= 16;

    // aliases :
    if ((t.si4==0)&&(t.snd==t.dst3)){
      if (t.name=="SUB"){
        t.name="NEG";
        t.instruction=" "+Y.registers_names[t.snd];
        if (t.snd == t.dst3)
          return 'FORM_R'
        t.instruction=" "+Y.registers_names[t.dst3];
        return 'FORM_RR'
      }
      if (t.name=="ORN"){
        t.name="NOT";
        t.instruction=" "+Y.registers_names[t.snd];
        if (t.snd == t.dst3)
          return 'FORM_R'
        t.instruction=" "+Y.registers_names[t.dst3];
       return 'FORM_RR'
      }
    }
    if ((f & Y.FORM_iR)&&(t.snd == t.dst3)) {
      t.instruction=" "+i4+" "+Y.registers_names[t.snd];
      return 'FORM_iR'
    }
    if(f & Y.FORM_Ri){ // PUT
      t.instruction=" "+Y.registers_names[t.snd]+" "+i4;
      return 'FORM_Ri'
    }
    // default
    if (f & Y.FORM_iRR) {

      if (t.snd == t.dst3) {
        t.instruction=" "+i4+" "+Y.registers_names[t.snd];
        return 'FORM_iR'
      }

      t.instruction=" "+i4+" "+Y.registers_names[t.snd]+" "+Y.registers_names[t.dst3];
      return 'FORM_iRR'
    }
  }
  else {
    if ((f & Y.FORM_RR)&&(t.snd == t.dst3)) {
      t.instruction=" "+Y.registers_names[t.si4]+" "+Y.registers_names[t.dst3];
      return 'FORM_RR'
    }
    if (f & Y.FORM_RRR) {
      t.instruction=" "+Y.registers_names[t.si4]+" "+Y.registers_names[t.snd]+" "+Y.registers_names[t.dst3];
      return 'FORM_RRR'
    }
  }
  add_message('red', 'Error : unknown or unhandled form ???');
 }

 this.decode_condition=function(t){
  var cond=(t.input & Y.FIELD_EXTCOND)>>Y.OFFSET_EXTCOND;
  if ((cond!=0) || (t.src4!=0)) { // exclude the ALWAYS condition
    var c=Y.tableCondition[cond]; // basename of the condition
    var o=Y.conditionTable[c]; // the condition object
    var n=Y.registers_names[t.src4]; // name of the register
    d=o[n]; // check if the register is associated with an alias
    if (typeof d=='string')
      t.instruction+=" "+d;
    else {
      t.instruction+=" "+c;
      if (o.reg != t.src4) // always, never
        t.instruction+=" "+n;
    }
  }   
 }


 // this.form is null => indicates the error.
 this.decode_instruction=function(input) {
  this.prefix=input & (Y.FIELD_IMMREG|Y.FIELD_EXTFORM);
  this.opcode=input &  Y.FIELD_OPCODE;
  this.input=input;
  this.instruction='';
  this.size_bytes=0;
  this.form=null;
  this.snd=(input & Y.FIELD_SND)>>Y.OFFSET_SND;
  this.si4=(input & Y.FIELD_SI4)>>Y.OFFSET_SI4;
  this.name=null;

  if (typeof input != 'number') {
    add_message('red', 'Error : wrong input format');
    return
  }

  // opcode lookup
  this.O=Y.opcode_table[this.opcode];

  if (typeof this.O != 'object') {
    if (input & Y.FIELD_IMMREG) {
      // output a DW
      add_message('red', 'Error : unknown opcode "'+int2hex(opcode)+'" (assuming a long instruction)');
      this.instruction = 'DW '+ int2hex(input_instruction,7,7);
      this.size_byte=4;
    }
    else {
      // output a DH
      add_message('red', 'Error : unknown opcode "'+int2hex(opcode)+'" (assuming a short instruction)');
      if (input & 0xFFFF0000)
         add_message('orange', 'Warning : garbage in the MSB');
      this.instruction = 'DH '+ int2hex(input_instruction & F4);
      this.size_byte=2;
    }
    return
  }

  this.name=this.O.name;
  var f=this.O.forms;
  var g=this.O.flags;

  // the size is straightforward to find :
  if (input & Y.FIELD_IMMREG) {
    // here we process a long instruction 
    this.size_bytes=4;

    // extended forms :
    if (input & Y.FIELD_EXTFORM) {
      this.dst3=(input>>Y.OFFSET_EXTDST3)&15;
      this.src4=(input>>Y.OFFSET_EXTSRC4)&15;
      this.form=this.decode_extended(this,f,g);
      this.decode_condition(this);
    }
    else { // Long Immediate instructions : FORM_RI, FORM_IR, FORM_IRR
      // test the garbage :
      if((input & 0xFF000000)&&(g&Y.IMM16_8LSB)
      || (input & 0xFFE00000)&&(g&Y.IMM16_5LSB))
        add_message('yellow', 'warning : garbage found in IMM16');
      this.form=this.decode_long(this,f,g);
    }
  }
  else { // Short forms
    this.size_bytes=2;
    if (f & Y.FORM_ALONE) {
      this.form='FORM_ALONE';
      if (this.snd != 0)
        add_message('yellow', 'warning : garbage found in SND');
      if (this.si4 != 0)
        add_message('yellow', 'warning : garbage found in SI4');
    }
    else
      this.form=this.decode_short(this,f,g);
  }

  // output
  if (this.form!=null){
    this.instruction=this.name+this.instruction;
    add_message('blue',' OK: opcode='+Y.decorate_opcode(this.opcode)+' instruction="<b>'+this.instruction+'</b>" ('+this.form+')');
  }
 }
}
